/*
   KallistiOS ##version##
   bubbles.c

   (c)2002 Dan Potter
*/

#include <kos.h>
#include <math.h>
#include <assert.h>
#include <plx/matrix.h>
#include <plx/prim.h>

#define prefetch(addr) asm("pref @%0" : : "r" (addr))

/*

This is a Parallax version of the "bubbles" KGL demo. The speed is about
the same, but it uses the Parallax functions instead of KGL.

*/

/* Draws a sphere out of GL_QUADS according to the syntax of glutSolidSphere

   This now uses triangle strips and PVR Direct Rendering for the best
   possible performance.

   Based very loosely on a routine from James Susinno.
*/
static float phase = 0.0f;
static pvr_poly_cxt_t cxt;
static pvr_poly_hdr_t hdr;

void mat_transform_sq(pvr_vertex_t *invecs, void *sq, int veccnt);

typedef struct {
	float radius;
	int slices;
	int stacks;
	pvr_vertex_t *data;
} sphere_t;

static void sphere(sphere_t *s) { /* {{{ */
	int	i, j;
	float	pitch, pitch2;
	float	r, g, b;
	float	yaw, yaw2;
	pvr_vertex_t *v;

	v = (pvr_vertex_t *)malloc(s->stacks * (s->slices+2) * sizeof(pvr_vertex_t) + 32);
	v = (pvr_vertex_t *)(((uint32)v & ~31) + 32); /* align to 32 bytes */
 	s->data = v;
	// s->data_trans = (pvr_vertex_t *)malloc(s->stacks * (s->slices+2) * sizeof(pvr_vertex_t));
	/* transformed data -- for testing mat_transform */
	printf("allocated %d bytes for %d stacks, %d + 2 slices, and %d-byte pvr_vertex_t\n",
			s->stacks * (s->slices+2) * sizeof(pvr_vertex_t), s->stacks, s->slices, sizeof(pvr_vertex_t));

	/* Iterate over stacks */
	for ( i=0; i<s->stacks; i++ ) {
		pitch = 2*M_PI * ( (float)i/(float)s->stacks );
		pitch2 = 2*M_PI * ( (float)(i+1)/(float)s->stacks );

		/* Iterate over slices: each entire stack will be one
		   long triangle strip. */
		for ( j=0; j<=s->slices/2; j++ ) {
			yaw = 2*M_PI * ( (float)j/(float)s->slices );
			yaw2 = 2*M_PI * ( (float)(j+1)/(float)s->slices );

			/* x, y+1 */
			v->x = s->radius * fcos( yaw ) * fcos( pitch2 );
			v->y = s->radius * fsin( pitch2 );
			v->z = s->radius * fsin( yaw ) * fcos( pitch2 );

			
			r = 0.75f;
			g = fcos( yaw*2 ) / 3.0f + 0.35f;
			b = 0.0f;
	
			v->argb = plx_pack_color(0.5f, r, g, b);
			v->flags = PVR_CMD_VERTEX;
			v++;


			/* x, y */
			v->x = s->radius * fcos( yaw ) * fcos( pitch );
			v->y = s->radius * fsin( pitch );
			v->z = s->radius * fsin( yaw ) * fcos( pitch );

			//r = 0.5f;
			r = 0.75f;
			g = fsin( yaw*2 ) / 3.0f + 0.35f;
			b = 0.0f;
			// b = fsin( phase + pitch ) / 2.0f + 0.5f;
			
			v->argb = plx_pack_color(0.5f, r, g, b);
			if (j == s->slices/2)
				v->flags = PVR_CMD_VERTEX_EOL;
			else
				v->flags = PVR_CMD_VERTEX;
			v++;
		}
	}
}

static void draw_sphere(sphere_t *s) {
	pvr_vertex_t *v;
	pvr_dr_state_t	dr_state;

	v = s->data;
	assert(v != NULL);

	/* Setup our Direct Render state: pick a store queue and setup QACR0/1 */
	pvr_dr_init(dr_state);

	/* Initialize xmtrx with the values from KGL */
	plx_mat_identity();
	plx_mat3d_apply(PLX_MAT_SCREENVIEW);
	plx_mat3d_apply(PLX_MAT_PROJECTION);
	plx_mat3d_apply(PLX_MAT_MODELVIEW);

	/* Put our own polygon header */
	pvr_prim(&hdr, sizeof(hdr));

	/* Transform and write vertices to the TA via the store queues */
	mat_transform_sq(v, (void *)0xe0000000, s->stacks * (s->slices + 2));
}

#define SPHERE_CNT 64
static float r = 0;
static void sphere_frame_opaque() {
	int i;
	uint64 start;
	static sphere_t big_sphere = { 1.2f, 20, 20, NULL };
	static sphere_t small_sphere = { 0.8f, 20, 20, NULL };

	if (!big_sphere.data)
		sphere(&big_sphere);
	if (!small_sphere.data)
		sphere(&small_sphere);

	vid_border_color(255, 0, 0);
	pvr_wait_ready();
	vid_border_color(255, 192, 0);
	start = timer_ms_gettime64();

	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -12.0f);
	plx_mat3d_rotate(r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(6.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 6.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		draw_sphere(&big_sphere);
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}
	plx_mat3d_pop();

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -12.0f);
	plx_mat3d_rotate(-r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(3.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 3.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		draw_sphere(&small_sphere);
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}
	plx_mat3d_pop();

	//printf("%d\n", (uint32)(timer_ms_gettime64() - start));

	vid_border_color(0, 0, 255);
	pvr_scene_finish();

	r++;
	phase += 2*M_PI / 240.0f;
}

static void sphere_frame_trans() {
	int i;

	vid_border_color(255, 0, 0);
	pvr_wait_ready();
	vid_border_color(0, 255, 0);

	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_TR_POLY);

	plx_mat3d_identity();
	plx_mat3d_translate(0.0f, 0.0f, -12.0f);
	plx_mat3d_rotate(r * 2, 0.75f, 1.0f, 0.5f);
	plx_mat3d_push();

	for (i=0; i<SPHERE_CNT; i++) {
		plx_mat3d_translate(4.0f * fcos(i * 2*M_PI / SPHERE_CNT), 0.0f, 4.0f * fsin(i * 2*M_PI / SPHERE_CNT));
		plx_mat3d_rotate(r, 1.0f, 1.0f, 1.0f);
		//sphere(1.0f, 20, 20);
		if (i < (SPHERE_CNT-1))
			plx_mat3d_peek();
	}

	plx_mat3d_pop();

	vid_border_color(0, 0, 255);
	pvr_scene_finish();

	r++;
	phase += 2*M_PI / 240.0f;
}

void do_sphere_test() {
	uint8		addr;
	cont_cond_t	cond;
	int		mode = 0;

	pvr_set_bg_color(0.2f, 0.05f, 0.0f);

	pvr_poly_cxt_col(&cxt, PVR_LIST_OP_POLY);
	cxt.gen.culling = PVR_CULLING_NONE;
	pvr_poly_compile(&hdr, &cxt);

	for (;;) {
		if (!mode)
			sphere_frame_opaque();
		else
			sphere_frame_trans();
	
		addr = maple_first_controller();
		if (addr == 0)
			continue;
		if (cont_get_cond(addr, &cond) < 0)
			continue;
		if (!(cond.buttons & CONT_START)) {
			return;
		}
	}
}

pvr_init_params_t params = {
	/* Enable opaque and translucent polygons with size 16 */
	{ PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
	
	/* Vertex buffer size */
	2560*1024
};
	
int main(int argc, char **argv) {
	/* Init PVR API */
	pvr_stats_t stats;

	if (pvr_init(&params) < 0)
		return -1;

//	cont_btn_callback(0, CONT_START, (cont_btn_callback_t)arch_exit);

	/* Init matrices */
	plx_mat3d_init();
	plx_mat3d_mode(PLX_MAT_PROJECTION);
	plx_mat3d_identity();
	plx_mat3d_perspective(45.0f, 640.0f / 480.0f, 0.1f, 100.0f);
	plx_mat3d_mode(PLX_MAT_MODELVIEW);

	/* Do the test */
	printf("Bubbles KGL sample: press START to exit\n");
	do_sphere_test();

	pvr_get_stats(&stats);
	dbglog(DBG_DEBUG, "3D Stats: %ld vblanks, frame rate ~%f fps, max vertex used %d bytes\n",
			stats.vbl_count, (double)stats.frame_rate, stats.vtx_buffer_used_max);

	return 0;
}

